"""Transports module."""
DEFAULT_TIMEOUT = 30
"""int: default timeout in seconds to use when interacting with the devices."""
DEFAULT_PORT = 22
"""int: default port to use when interacting with the devices."""
